#!/usr/bin/env python
"""
Program for generating skeleton files for C/C++ projects.
"""

import codecs
from optparse import OptionParser
from mako.runtime import Context
from mako.lookup import TemplateLookup
import pwd
import os
import glob
import time
import sys
import string

"""Getting a single character code is taken from:
http://code.activestate.com/recipes/134892/
"""
class _Getch:
    """Gets a single character from standard input.  Does not echo to the
screen."""
    def __init__(self):
        try:
            self.impl = _GetchWindows()
        except ImportError:
            self.impl = _GetchUnix()

    def __call__(self): return self.impl()


class _GetchUnix:
    def __init__(self):
        import tty, sys

    def __call__(self):
        import sys, tty, termios
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(sys.stdin.fileno())
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch


class _GetchWindows:
    def __init__(self):
        import msvcrt

    def __call__(self):
        import msvcrt
        return msvcrt.getch()

getch = _Getch()


def check_outfile(outname):
    k='n'
    if os.path.exists(outname):
        msg = "File %s exists [s/S skip; o/O overwrite]:" % outname
        print msg 
        inkey = _Getch()
        for i in xrange(sys.maxint):
            k=inkey()
            if k in ['s','S','o','O']:
                break
            print msg 
    if k == 'O':
        k = 'o'
    if k == 'S':
        k = 's'
    return k

def setup_context(cstopt, noquestions=False):
    """Prompt user for various customisation informations"""
    userentry = pwd.getpwuid(os.getuid())
    author = userentry[4].split(',')[0]
    login = userentry[0]
    fhandle = open('/etc/mailname')
    if fhandle :
        hostmail = fhandle.readline().strip()
    email = login+'@'+hostmail
    license = 'BSD'
    if cstopt['femdk']:
        owner = 'Cracow University of Technology'
    else:
        owner = author

    organisation = owner
    short_description = "Ask the program author what is it purpose"

    date = time.strftime("%c")
    year = time.strftime("%Y")

    if not noquestions:
        inpt = raw_input("One line program description for usage string?  \n")
        if inpt.strip() != "":
            short_description = inpt
        inpt = raw_input("What is the name of the author? [%s]\n" % author)
        if inpt.strip() != "":
            author = inpt
        inpt = raw_input("What is the email address of the author? [%s]\n" % 
                    email)
        if inpt.strip() != "":
            email = inpt
        inpt = raw_input("What is the license "\
                         "(possible BSD, GPL, LGPL)? [%s]\n" % license)
        if inpt.strip() != "":
            license = inpt
        inpt = raw_input("What is the name of copyright holder? "\
                         "[%s]\n" % owner)

    cstopt['author'] = author
    cstopt['email'] = email
    cstopt['date'] = date
    cstopt['license'] = license
    cstopt['owner'] = owner
    cstopt['year'] = year
    cstopt['organisation'] = organisation
    cstopt['short_description'] = short_description
 
def header_guard(basename, prefix=''):
    """Construct header guard from the base name"""
    guard = basename.upper()
    guard=prefix.upper() + '_' + guard + '_HXX'
    return guard

def generate_module(basename, options, custom_options):
    """Generate file for a C++ module: header file and source file.
    """
    if options.femdk:
        prefix='femdk'
    else:
        prefix=''

    custom_options['header_guard'] = header_guard(basename, prefix)
    custom_options['header_name'] = basename+".hxx"

    header_template = mylookup.get_template('mkcxx_header.mako')
    source_template = mylookup.get_template('mkcxx_source.mako')

    header_output  = codecs.open(basename+".hxx", "w", "iso-8859-1")
    source_output  = codecs.open(basename+".cxx", "w", "iso-8859-1")

    ctx = Context(header_output, **custom_options)
    header_template.render_context(ctx)

    ctx = Context(source_output, **custom_options)
    source_template.render_context(ctx)

    generate_bootstrap(basename, options, custom_options)
    generate_cleanALL(basename, options, custom_options)

def generate_qt_classes(basename, options, custom_options):
    """Generate sourc files for a Qt based application.
    """

    custom_options['header_guard'] = header_guard('EXCEPTION',
            options.classPrefix)

    except_template = mylookup.get_template('mkcxx_qt_except_h.mako')
    except_output = codecs.open(options.classPrefix+'Exception.h', 'w',
                                'iso-8859-1');
    ctx = Context(except_output, **custom_options)
    except_template.render_context(ctx)

def generate_test(basename, options, custom_options):
    """Generate source file for a C++ test suite.
    """
    if options.femdk:
        prefix='femdk'
    else:
        prefix=''

    test_template = mylookup.get_template('mkcxx_test_source.mako')

    test_output  = codecs.open(basename+".cxx", "w", "iso-8859-1")

    ctx = Context(test_output, **custom_options)

    test_template.render_context(ctx)

def generate_bootstrap(basename, options, custom_options):
    """Generate bootstrap script.
    """
    if options.femdk:
        prefix='femdk'
    else:
        prefix=''
    path = os.path.join(options.prefix, "build", "bootstrap")

    script_template = mylookup.get_template('mkcxx_bootstrap.mako')

    script_output  = codecs.open(path, "w", "iso-8859-1")

    ctx = Context(script_output, **custom_options)
    script_template.render_context(ctx)

def generate_cleanALL(basename, options, custom_options):
    """Generate cleanALL script.
    """
    path = os.path.join(options.prefix, "build", "cleanALL")

    script_template = mylookup.get_template('mkcxx_cleanALL.mako')

    script_output  = codecs.open(path, "w", "iso-8859-1")

    ctx = Context(script_output, **custom_options)
    script_template.render_context(ctx)

def generate_program(basename, options, custom_options):
    """Generate file for a C++ program.
    """
    if options.femdk:
        prefix='femdk'
    else:
        prefix=''

    custom_options['header_guard'] = header_guard(basename, prefix)

    program_template = mylookup.get_template('mkcxx_program_source.mako')
    conf_template = mylookup.get_template('mkcxx_program_conf.mako')

    program_output  = codecs.open(basename+".cxx", "w", "iso-8859-1")
    conf_output = codecs.open(basename+'_config.h.in', 'w', "iso-8859-1") 

    ctx = Context(program_output, **custom_options)
    program_template.render_context(ctx)

    ctx = Context(conf_output, **custom_options)
    conf_template.render_context(ctx)

    generate_builder(basename, options, custom_options)

    generate_cmdlparser(basename, options, custom_options)

    generate_bootstrap(basename, options, custom_options)
    generate_cleanALL(basename, options, custom_options)

    if options.qt :
        generate_qt_classes(basename, options, custom_options)

def generate_builder(basename, options, custom_options):
    """Generate builder file, i.e Makefile, CMakeLists.txt, etc.
    """
    
    buildtype = options.buildtype;

    if buildtype == 'none':
        return
    elif buildtype == 'autotools':
        print "WARNING: autotools builder is not supported yet."\
              " Skipping the request"
        return

    template_name = 'mkcxx_' + options.gentype + '_' + options.buildtype \
                    + '.mako'
    builder_template = mylookup.get_template(template_name)

    outnamemap = {'cmake' : 'CMakeLists.txt',
                  'make' : 'Makefile',
                 }
    outname = outnamemap[options.buildtype]

    nso = check_outfile(outname);

    if nso != 's':
        builder_output  = codecs.open(outname, "w", "iso-8859-1")

        ctx = Context(builder_output, **custom_options)
        builder_template.render_context(ctx)


def generate_cmdlparser(basename, options, custom_options):
    """Generate command line parser.
    """

    cmdlgener = options.cmdlgener;

    if cmdlgener == 'none':
        return

    template_name = 'mkcxx_program_optdef.mako'
    cmdlparser_template = mylookup.get_template(template_name)

    outname = basename + '_options.def'

    cmdlparser_output  = codecs.open(outname, "w", "iso-8859-1")

    ctx = Context(cmdlparser_output, **custom_options)
    cmdlparser_template.render_context(ctx)

def setup_parser(parser, generators, buildtypes, cmdlgeners):
    """Define options for command line parser.
    """
    cmdlparser.add_option("--femdk-install-dir", metavar='DIR',
                          dest='femdkdir',
                          help="FEMDK installation directory")

    cmdlparser.add_option("--cmake-link-dir", metavar='DIR',
                          dest='cmakedir',
                          help="Directory from which to link CMAKE modules")

    cmdlparser.add_option("--template-dir", metavar="DIR", dest="tdir", 
                         help ="Search for template files in DIR directory", 
                         )

    cmdlparser.add_option("--femdk", action="store_true", dest="femdk",
                          help ="Generate skeleton file for FEMDK project"
                         )
    cmdlparser.add_option("--qt", action="store_true", dest="qt",
                          help = "Generate classes for Qt based program"
                        )
    cmdlparser.add_option("--no-femdk-main", action="store_false",
                          dest="femdk_main",
                          help ="Do not use main via STANDARD_FEMDK_MAIN macro"
                         )

    cmdlparser.add_option("--noquestions", action="store_true", 
                          dest="noquestions",
                          help = "Do not ask any questions"
                         )

    gentypes = string.join(map(lambda x : "'%s'"%x, generators.keys()),',')

    cmdlparser.add_option("--type", type="choice", choices=generators.keys(),
                          dest="gentype", 
                          help = """Type of output to generate.
                          Possible types: [ %s ]""" % gentypes 
                         )

    bt = string.join(map(lambda x : "'%s'"%x, buildtypes),',')
    cmdlparser.add_option("--build-type", type="choice", choices=buildtypes,
                          dest="buildtype",
                          help = """Build system to use. Possible build types:
 [ %s ]""" % bt
                         )
    cmdl = string.join(map(lambda x : "'%s'"%x, cmdlgeners),',')
    cmdlparser.add_option("--cmdl-generator", type="choice", choices=cmdlgeners,
                          dest="cmdlgener",
                          help = """Command line parser generator to use.
                          Possible generators: [ %s ]""" % cmdl 
                         )

    cmdlparser.add_option("--nodir", action="store_true", 
                          dest="nodir",
                          help = "Do not create output directory"
                         )

    cmdlparser.add_option("--prefix", type='string', 
                          dest="prefix",
                          help = """Specify the prefix to be added to output
directory path [prefix/basename]"""
                         )

    cmdlparser.add_option("--test-class", type='string', 
                          dest="testClass",
                          help = """Name of the test suite class. If not given 'basename' is used."""
                         )

    cmdlparser.add_option("--class-prefix", type='string',
                          dest="classPrefix",
                        help = """Prefix added to class name. Default empty."""
                         )

    cmdlparser.set_defaults(buildtype = 'cmake',
                            gentype = 'module',
                            noquestions = False,
                            femdk = False,
                            qt = False,
                            femdkdir = "/usr",
                            cmakedir = "",
                            tdir = '%TEMPLATEDIR%',
                            nodir = False,
                            prefix='./',
                            cmdlgener = 'autoopt',
                            classPrefix=''
                            )

def setup_output_dir(basename, options):
    """Create directory for output files and set is as current.
    """
    path = options.prefix
    if options.nodir is False:
        path = os.path.join(options.prefix, basename) 
        try:
            os.makedirs(path)
        except OSError:
            print "ERROR: The path: '%s' already exists" % path
            print "Either remove or rename it and try again"
            sys.exit(1)
    os.chdir(path)
    if options.gentype in  ['program', 'module']:
        os.makedirs('config')
        os.makedirs('config/cmake_modules')
        from os import listdir
        from os.path import isfile, join

        if options.cmakedir == "":
          mypath=os.path.join(options.femdkdir, "share/femdk/cmake_modules")
        else:
          mypath=options.cmakedir
 
        cmakefiles = glob.glob(mypath + "/*.cmake")
        for f in cmakefiles:
            if options.cmakedir == "":
              os.symlink(f, os.path.join("config/cmake_modules",
                  os.path.basename(f)))
            else:
              os.symlink(os.path.relpath(f, "./config/cmake_modules"), 
                  os.path.join("config/cmake_modules", os.path.basename(f)))

    if options.gentype == 'program':
        os.makedirs('options_def')
    os.makedirs('build')

if __name__ == '__main__':

    usage = """%prog [options] basename\n\n""" \
    """Generate skeleton source files for different types of source: program
source, library source and header, test suite source, build scripts, etc. """

    generators = { 'module' : generate_module,
                   'program' : generate_program,
                   'test' : generate_test
                 }
    buildtypes = ['none', 'cmake', 'make', 'autotools']
    
    cmdlgeners = ['none', 'autoopt']

    cmdlparser = OptionParser(usage=usage, version="%prog %VERSION%")

    setup_parser(cmdlparser, generators, buildtypes, cmdlgeners)

    (options, args) = cmdlparser.parse_args()

    if len(args) < 1:
        print "ERROR: missing skeleton base name\n"
        print usage
        sys.exit(1)

    basename = args[0]

    dirs = [] # directories for template search
    dirs.append(options.tdir)
    mylookup = TemplateLookup(
            output_encoding = 'iso-8859-1',
            directories=dirs)

    custom_options = {}

    custom_options['femdk'] = options.femdk
    custom_options['qt'] = options.qt
    custom_options['femdkinstalldir'] = options.femdkdir
    custom_options['testClass'] = options.testClass
    custom_options['basename'] = basename
    custom_options['cmdlgener'] = options.cmdlgener
    custom_options['qtcp'] = options.classPrefix

    setup_output_dir(basename, options)

    setup_context(custom_options, options.noquestions)

    # Generate output
    generators[options.gentype](basename, options, custom_options)
